/*
CHGDIR.C
Win32 Console app
Demonstrates that Win32 tasks in Win95 keep current drv/dir in Win16 TDB

Andrew Schulman
andrew@ora.com
http://www.ora.com/windows/
ftp://ftp.ora.com/pub/examples/windows/win95.update/schulman.html
August 1995

This version is corrected from the version that appears in my book
*Unauthorized Windows 95* (IDG Books, 1994), pp. 446-448:

After the book was written, the GetCurrentThreadId() function in
Win95 was changed to return obfuscated thread IDs.  CHGDIR.C now uses
the TIDToTDB() function from UNOBFUSC.C to de-obfuscate the thread
ID into a Ring 3 thread control block.

Also after the book was written, KERNEL32.DLL stopped exporting the
VxDCall() function by name.  This function has ordinal #1 in KERNEL32.DLL.
However, KERNEL32 does not allow imports by ordinal.  To access VxDCall,
CHGDIR.C now uses the GetK32ProcAddress() function from K32EXP.C.

cl chgdir.c unobfusc.c k32exp.c

For instructions on using this program, see *Unauthorized Windows
95*, p. 450.

Sample session with CHGDIR (see numbered comments below):

C:\web>chgdir
Thread ID: FFFD6F4Fh
TDB: 8152E478
TDB: 329Eh
TDB base: 801AB580h                 ;;; (1)
PSP: 3287h

[C:\web] md foo
[c:\web] cd foo
TDB+66h:  82h                       ;;; (2)
TDB+100h: '\web\foobar'
[C:\web\foobar] md This is a long directory name
[C:\web\foobar] cd This is a long directory name
TDB+66h:  82h
TDB+100h: '\web\foobar\This is a long directory name'       ;;; (3)
[C:\web\foobar\This is a long directory name] cd k:         ;;; (4)
TDB+66h:  8Ah
TDB+100h: '\'
[K:\] md Yet another directory on a different drive
[K:\] cd Yet another directory on a different drive
TDB+66h:  8Ah
TDB+100h: '\Yet another directory on a different drive'
[K:\Yet another directory on a different drive] cd c:
TDB+66h:  82h
TDB+100h: '\web'                                            ;;; (5)
[C:\web] exit
done

A few comments on the above:  

(1) The 32-bit CHGDIR app can directly access the 16-bit TDB, using its
    base address.  CHGDIR gets the base by calling a DPMI function via
    the Win32 VxDCall function.
        
(2) TDB stores the task's current drive.  82h = C:, 8Ah = K:

(3) TDB stores the task's current directory, including directories with
    long (LFN) names, with embedded spaces, etc.
        
(4) The CD command in CHGDIR is passed directly through to the 
    SetCurrentDirectory() Win32 API call, which in turn allows changing
    drive letters.
        
(5) Whoops!  SetCurrentDirectory() from K: to C: didn't bring us back to
    the most recent current directory this task had on C:
*/

#include <stdlib.h>
#include <stdio.h>
#define  WIN32_LEAN_AND_MEAN
#include "windows.h"
#include "unobfusc.h"
#include "k32exp.h"

DWORD (WINAPI *VxDCall)(DWORD srvc, DWORD eax, DWORD ecx);

#define GET_PROC(mod, func) GetProcAddress(GetModuleHandle(mod), (func))
#define VWIN32_INT21_CALL   0x2A0010
#define VWIN32_INT31_CALL   0x2A0029
#define DosCall(eax, ecx)   VxDCall(VWIN32_INT21_CALL, (eax), (ecx))
#define DPMICall(eax, ecx)  VxDCall(VWIN32_INT31_CALL, (eax), (ecx))

char curdir[MAX_PATH], buf[MAX_PATH], bufupr[MAX_PATH];

void fail(const char *s) { puts(s); exit(1); }

main()
{
    HANDLE id;
    BYTE *tdb_drv, *tdb_curdir, *tdb_base;
    WORD tdb, psp;
    
    // Thread Id == OBFUSCATED Ring 3 THCB (Thread Control Block)
    id = GetCurrentThreadId();
    printf("Thread ID: %08lXh\n", id);
    
    // thread id unobfuscator from UNOBFUSC.C
    if ((id = TIDToTDB(id)) == 0)
        fail("Couldn't find thread ID unobfuscator XOR pattern");
    
    printf("TDB: %08lX\n", id);
        
    // Win16 TDB is at offset 1Ch in Ring 3 THCB
    // I'm a little surprised it's so easy to get at this!
    tdb = *((WORD *) (((BYTE *) id) + 0x1c));
    printf("TDB: %04Xh\n", tdb);
    
#if 1
    if ((VxDCall = GetK32ProcAddress(1)))               // see K32EXP.C
#else
    // Use run-time dynamic linking to turn strings
    // into callable function pointers 
    if ((VxDCall = GET_PROC("KERNEL32", "VxDCall0")))  // not any more!
#endif      
    {
        // Get linear base address for TDB by calling
        // DPMI INT 31h function 6 (Get Selector Base).
        // Even though we're a 32-bit app, we get the 16-bit
        // DPMI services, since we're calling indirectly.
        _asm mov bx, tdb
        DPMICall(0x0006, 0);
        _asm mov word ptr tdb_base+2, cx
        _asm mov word ptr tdb_base, dx
        printf("TDB base: %08lXh\n", tdb_base);
        
        // Call DOS INT 21h function 62h (Get PSP)
        DosCall(0x6200, 0);
        _asm mov psp, bx
        printf("PSP: %04Xh\n\n", psp);

        // Sanity check:  is TDB[60h] == PSP?
        if (*((WORD *) (tdb_base + 0x60)) != psp)
            fail("TDB and PSP don't match!");
            
        // Current drv at offset 66h in TDB, curr dir at offset 100h
        // SetCurrentDirectory changes these.
        // Note: Current drv/dir on per-task basis, not per-thread.
        tdb_curdir = tdb_base + 0x100;
        tdb_drv = tdb_base + 0x66;
    }
    else
        printf("Warning: Can't access VxDCall0\n\n");
    
    for (;;)
    {
        if (GetCurrentDirectory(MAX_PATH, curdir) == 0)
            strcpy(curdir, "invalid");
        printf("[%s] ", curdir);
        gets(buf);
        if ((buf[0] == '\0') || (buf[0] == '\n')) 
            continue;
        strcpy(bufupr, buf);
        strupr(bufupr);
        #define MATCH(cmd)  (strncmp(bufupr, (cmd), sizeof(cmd)-1) == 0)
        if (MATCH("EXIT"))
            break;
        else if (MATCH("CD "))
        {
            if (! SetCurrentDirectory(&buf[3]))
                printf("CD failed\n");
            else if (tdb_curdir)
            {
                printf("TDB+66h:  %02Xh\n", *tdb_drv);
                printf("TDB+100h: \'%s\'\n", tdb_curdir);
            }
        }
        else if (MATCH("MD "))
        {
            if (! CreateDirectory(&buf[3], 0))
                printf("MD failed\n");
        }
        else
            WinExec(buf, SW_NORMAL);
    }
    printf("done\n");
    return 0;
}
